/*{{{  Header*/
/*------------------------------------------------------------------*/
/*                                                                  */
/*                            socket.c                              */
/*                                                                  */
/*     by Nick Clarke, Southampton University Computing Services    */
/*                                                                  */
/*     Functions added to iserver in order to implement a BSD 4.3   */
/*     socket library on the root transputer.                       */
/*                                                                  */
/*     Tag choices correspond to Inmos B300 (TCPLINK) tags.         */
/*                                                                  */
/*------------------------------------------------------------------*/
/*}}}*/
/*{{{F isocket.h*/
/*:::F isocket.h*/
/*}}}*/
/*{{{  includes*/
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <netdb.h>
#include <pwd.h>

#include <stdio.h>
#include <errno.h>

#include <dirent.h>

#include "server.h"
#include "iserver.h"
#include "pack.h"

#include "isocket.h"
/*}}}*/
/*{{{  globals*/
extern bool CocoPops;
extern bool VerboseSwitch;
bool OutSigio;

extern unsigned char Tbuf[TRANSACTION_BUFFER_SIZE];

extern int TheLink;

typedef struct seg {
  int  Length;
  int  Done;
  unsigned char Buff[4];    /* Break array bounds for larger buffers */
  } Segment;
/*}}}*/

/*{{{  SigioHandler*/
void SigioHandler()
  {
    OutSigio=OutSigio | 1;
  }
/*}}}*/
/*{{{  SigurgHandler*/
void SigurgHandler()
  {
    OutSigio=OutSigio | 2;
  }
/*}}}*/
void SpSocket()
{
  unsigned char Command;

   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
    Command =get_8();
   switch (Command)
    {
    /*{{{  all the possible socket commands*/
    /*{{{  accept()*/
    case SP_ACCEPT :
      { long SockId, MaxAddrLen, AddrLen, SockErr, NewSockId;
        unsigned char Result, AddrSpace[MAXSTRINGLEN];
        struct sockaddr *Addr=(struct sockaddr *)AddrSpace;
        short Family;
    
        dbgmsg( "SP.ACCEPT" );
         SockId = get_32(); AddrLen  = get_32();
        dbgmsg( "Socket id %d, Max Addr Len %d", SockId, AddrLen );
        MaxAddrLen=AddrLen;
        /*{{{  do accept, sort out error returns*/
        if ((NewSockId=accept(SockId,Addr,&AddrLen))<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else if (MaxAddrLen<AddrLen)
          {
            SockErr=EFAULT;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          {
            put_32( NewSockId );
            Family=Addr->sa_family;
            put_16( Family );
            put_slice( AddrLen, AddrSpace );
          }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  bind()*/
    case SP_BIND :
      { long SockId, AddrLen, SockErr;
        unsigned char Result, AddrSpace[MAXSTRINGLEN];
        struct sockaddr *Addr=(struct sockaddr *)AddrSpace;
        short Family;
    
        dbgmsg( "SP.BIND" );
         SockId  = get_32();
        dbgmsg( "Socket id %d", SockId );
         Family  = get_16();
        dbgmsg( "Family %d",(int)Family );
        AddrLen = get_slice( AddrSpace ); Addr->sa_family=Family;
        dbgmsg( "Got %d address bytes",AddrLen );
        /*{{{  do bind, sort out error returns*/
        if (bind(SockId,Addr,AddrLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  close()*/
    case SP_SOCKCLOSE :
      { long SockId, SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.SOCKET.CLOSE" );
         SockId  = get_32();
        dbgmsg( "Socket id %d", SockId );
        /*{{{  do close, sort out error returns*/
        if (close(SockId)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  inet_addr()*/
    case SP_INETADDR :
      { long SockId, Flags, Mode, SockErr, PathLen;
        unsigned char Result, Path[MAXSTRINGLEN+1];
    
        dbgmsg( "SP.INETADDR" );
        PathLen = get_slice( Path );
        Path[PathLen<MAXSTRINGLEN ? PathLen : MAXSTRINGLEN]=(char)'\0';
        dbgmsg( "Test address %s ", Path );
        /*{{{  do inet_addr, sort out error returns*/
        if ((SockId=inet_addr(Path))!=-1)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
        {
          put_32( SockId );
          dbgmsg( "Address %d", SockId );
        }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  connect()*/
    case SP_CONNECT :
      { long SockId, AddrLen, SockErr;
        unsigned char Result, AddrSpace[MAXSTRINGLEN];
        struct sockaddr *Addr=(struct sockaddr *)AddrSpace;
        short Family;
    
        dbgmsg( "SP.CONNECT" );
         SockId  = get_32();
        dbgmsg( "Socket id %d", SockId );
         Family  = get_16();
        dbgmsg( "Family %d",(int)Family );


        AddrLen = get_slice( AddrSpace ); Addr->sa_family=Family;
        dbgmsg( "Got %d address bytes",AddrLen );


        /*{{{  do connect, sort out error returns*/
        if (connect(SockId,Addr,AddrLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  gethostbyname()*/
    case SP_GETHOSTBYNAME :
      { long MaxHostentLen, SockErr=0, NameLen;
        unsigned char Result, Name[MAXSTRINGLEN], Space[MAXSTRINGLEN];
        struct hostent *Hostent;
    
        dbgmsg( "SP.GETHOSTBYNAME" );
        NameLen = get_slice( Name );
         MaxHostentLen  = get_32();
        *(Name+MAXSTRINGLEN-1)=(char)0;        /* ensure it is terminated */
        dbgmsg( "Host >%s<",Name );
        if ((Hostent=gethostbyname(Name))==NULL) Result=ER_ERROR;
        else Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        /*{{{  transfer data pointed at by Hostent into buffer*/
        if (Result==ER_SUCCESS)
        { int Nalias, Naddr, Len, i;
          char *p;
          for(Nalias=0;((Hostent->h_aliases[Nalias])!=NULL);Nalias++) ;
          put_32( Nalias );
          for(Naddr=0;((Hostent->h_addr_list[Naddr])!=NULL);Naddr++) ;
          put_32( Naddr );
          Len=(Hostent->h_name!=NULL) ? strlen(Hostent->h_name)+1 : 0;
          put_slice( Len, (Hostent->h_name) );
          for(i=0;i<Nalias;i++) {
            p=Hostent->h_aliases[i];
            Len=strlen(p)+1;
            put_slice( Len, p );
            }
          i=Hostent->h_addrtype; put_32( i );
          i=Hostent->h_length; put_32( i );
          for(i=0;i<Naddr;i++) {
            p=Hostent->h_addr_list[i];
            Len=Hostent->h_length;
            put_slice( Len, p );
            }
        }
        /*}}}*/
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  gethostbyaddr()*/
    case SP_GETHOSTBYADDR :
      { long MaxHostentLen, SockErr=0, AddrLen, Type;
        unsigned char Result, Addr[MAXSTRINGLEN], Space[MAXSTRINGLEN];
        struct hostent *Hostent;
    
        dbgmsg( "SP.GETHOSTBYADDR" );
        AddrLen = get_slice( Addr );
         Type  = get_32();
         MaxHostentLen  = get_32();
        *(Addr+MAXSTRINGLEN-1)=(char)0;        /* ensure it is terminated */
        dbgmsg( "Address %d, Type %d",Addr,Type );
        if ((Hostent=gethostbyaddr(Addr,AddrLen,Type))==NULL) Result=ER_ERROR;
        else Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        /*{{{  transfer data pointed at by Hostent into buffer*/
        if (Result==ER_SUCCESS)
        { int Nalias, Naddr, Len, i;
          char *p;
          for(Nalias=0;((Hostent->h_aliases[Nalias])!=NULL);Nalias++) ;
          put_32( Nalias );
          for(Naddr=0;((Hostent->h_addr_list[Naddr])!=NULL);Naddr++) ;
          put_32( Naddr );
          Len=(Hostent->h_name!=NULL) ? strlen(Hostent->h_name)+1 : 0;
          put_slice( Len, (Hostent->h_name) );
          for(i=0;i<Nalias;i++) {
            p=Hostent->h_aliases[i];
            Len=strlen(p)+1;
            put_slice( Len, p );
            }
          i=Hostent->h_addrtype; put_32( i );
          i=Hostent->h_length; put_32( i );
          for(i=0;i<Naddr;i++) {
            p=Hostent->h_addr_list[i];
            Len=Hostent->h_length;
            put_slice( Len, p );
            }
        }
        /*}}}*/
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  gettimeofday()*/
    case SP_GETTIMEOFDAY :
      { long MaxHostentLen, SockErr=0, AddrLen, Type;
        unsigned char Result, Addr[MAXSTRINGLEN], Space[MAXSTRINGLEN];
        struct timeval tp; struct timezone tzp;
    
        dbgmsg( "SP.GETTIMEOFDAY" );
         (tp.tv_sec)  = get_32();
         (tp.tv_usec)  = get_32();
         (tzp.tz_minuteswest)  = get_32();
         (tzp.tz_dsttime)  = get_32();
        if (gettimeofday(&tp,&tzp)<0) Result=ER_ERROR;
        else Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
        {
          put_32( (tp.tv_sec) );
          put_32( (tp.tv_usec) );
          put_32( (tzp.tz_minuteswest) );
          put_32( (tzp.tz_dsttime) );
        }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  sethostent()*/
    case SP_SETHOSTENT :
      { long StayOpen, SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.SETHOSTENT" );
         StayOpen  = get_32();
        sethostent(StayOpen);
        Result=ER_SUCCESS;
        SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  endhostent()*/
    case SP_ENDHOSTENT :
      { long SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.ENDHOSTENT" );
        endhostent();
        Result=ER_SUCCESS;
        SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  gethostname()*/
    case SP_GETHOSTNAME :
      { long MaxNameLen, SockErr;
        unsigned char Result, Name[MAXSTRINGLEN];
    
        dbgmsg( "SP.GETHOSTNAME" );
         MaxNameLen  = get_32();
        /*{{{  do gethostname, sort out error returns*/
        if (gethostname(Name,MaxNameLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        *(Name+MAXSTRINGLEN-1)=(char)0;        /* ensure it is terminated */
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          { long tmplen=strlen(Name)+1;
            put_slice( tmplen, Name );
          }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  sethostname()*/
    case SP_SETHOSTNAME :
      { long NameLen, SockErr;
        unsigned char Result, Name[MAXSTRINGLEN];
    
        dbgmsg( "SP.SETHOSTNAME" );
        NameLen = get_slice( Name );
        /*{{{  do sethostname, sort out error returns*/
        if (sethostname(Name,NameLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  getnetbyname()*/
    case SP_GETNETBYNAME :
      { long MaxNetentLen, SockErr=0, NameLen;
        unsigned char Result, Name[MAXSTRINGLEN], Space[MAXSTRINGLEN];
        struct netent *Netent;
    
        dbgmsg( "SP.GETNETBYNAME" );
        NameLen = get_slice( Name );
         MaxNetentLen  = get_32();
        *(Name+MAXSTRINGLEN-1)=(char)0;        /* ensure it is terminated */
        dbgmsg( "Network >%s<",Name );
        if ((Netent=getnetbyname(Name))==NULL) Result=ER_ERROR;
        else Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        /*{{{  transfer data pointed at by Netent into buffer*/
        if (Result==ER_SUCCESS)
        { int Nalias, Len, i;
          char *p;
          for(Nalias=0;((Netent->n_aliases[Nalias])!=NULL);Nalias++) ;
          put_32( Nalias );
          Len=(Netent->n_name!=NULL) ? strlen(Netent->n_name)+1 : 0;
          put_slice( Len, (Netent->n_name) );
          for(i=0;i<Nalias;i++) {
            p=Netent->n_aliases[i];
            Len=strlen(p)+1;
            put_slice( Len, p );
            }
          i=Netent->n_addrtype; put_32( i );
          i=Netent->n_net; put_32( i );
        }
        /*}}}*/
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  getnetbyaddr()*/
    case SP_GETNETBYADDR :
      { long MaxNetentLen, SockErr=0, Net, Type;
        unsigned char Result, Space[MAXSTRINGLEN];
        struct netent *Netent;
    
        dbgmsg( "SP.GETNETBYADDR" );
         Net  = get_32();
         Type  = get_32();
         MaxNetentLen  = get_32();
        dbgmsg( "Address >%d<",Net );
        if ((Netent=getnetbyaddr(Net,Type))==NULL) Result=ER_ERROR;
        else Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        /*{{{  transfer data pointed at by Netent into buffer*/
        if (Result==ER_SUCCESS)
        { int Nalias, Len, i;
          char *p;
          for(Nalias=0;((Netent->n_aliases[Nalias])!=NULL);Nalias++) ;
          put_32( Nalias );
          Len=(Netent->n_name!=NULL) ? strlen(Netent->n_name)+1 : 0;
          put_slice( Len, (Netent->n_name) );
          for(i=0;i<Nalias;i++) {
            p=Netent->n_aliases[i];
            Len=strlen(p)+1;
            put_slice( Len, p );
            }
          i=Netent->n_addrtype; put_32( i );
          i=Netent->n_net; put_32( i );
        }
        /*}}}*/
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  setnetent()*/
    case SP_SETNETENT :
      { long StayOpen, SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.SETNETENT" );
         StayOpen  = get_32();
        setnetent(StayOpen);
        Result=ER_SUCCESS;
        SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  endnetent()*/
    case SP_ENDNETENT :
      { long SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.ENDNETENT" );
        endhostent();
        Result=ER_SUCCESS;
        SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  getpeername()*/
    case SP_GETPEERNAME :
      { long SockId, MaxAddrLen, AddrLen, SockErr;
        unsigned char Result, AddrSpace[MAXSTRINGLEN];
        struct sockaddr *Addr=(struct sockaddr *)AddrSpace;
        short Family;
    
        dbgmsg( "SP.GETPEERNAME" );
         SockId = get_32(); AddrLen  = get_32();
        dbgmsg( "Socket id %d, Max Addr Len %d", SockId, AddrLen );
        /*{{{  do getpeername, sort out error returns*/
        if (getpeername(SockId,Addr,&AddrLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          {
            Family=Addr->sa_family;
            put_16( Family );
            put_slice( AddrLen, AddrSpace );
          }
        put_count(OutCount); return;
      }
    /*}}}*/
    case SP_GETPROTOENT :
    case SP_GETPROTOBYNAME :
    case SP_GETPROTOBYNUM :
    /*{{{  setprotoent()*/
    case SP_SETPROTOENT :
      { long StayOpen, SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.SETPROTOENT" );
         StayOpen  = get_32();
        setprotoent(StayOpen);
        Result=ER_SUCCESS;
        SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  endprotoent()*/
    case SP_ENDPROTOENT :
      { long SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.ENDPROTOENT" );
        endprotoent();
        Result=ER_SUCCESS;
        SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  getsockname()*/
    case SP_GETSOCKNAME :
      { long SockId, AddrLen, SockErr;
        unsigned char Result, AddrSpace[MAXSTRINGLEN];
        struct sockaddr *Addr=(struct sockaddr *)AddrSpace;
        short Family;
    
        dbgmsg( "SP.GETSOCKNAME" );
         SockId = get_32(); AddrLen  = get_32();
        dbgmsg( "Socket id %d, Max Addr Len %d", SockId, AddrLen );
        /*{{{  do getsockname, sort out error returns*/
        if (getsockname(SockId,Addr,&AddrLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          {
          Family=Addr->sa_family;
          put_16( Family );
          put_slice( AddrLen, AddrSpace );
          }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  getsockopt()*/
    case SP_GETSOCKOPT :
      { long SockId, Level, OptName, OptValLen, SockErr;
        unsigned char Result;
        char OptVal[MAXSTRINGLEN];
    
        dbgmsg( "SP.GETSOCKOPT" );
         SockId = get_32(); Level  = get_32();
         OptName = get_32(); OptValLen  = get_32();
        dbgmsg( "Socket id %d, Level %d", SockId, Level );
        /*{{{  do getsockopt, sort out error returns*/
        if (getsockopt(SockId,Level,OptName,OptVal,&OptValLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
            put_slice( OptValLen, OptVal );
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  setsockopt()*/
    case SP_SETSOCKOPT:
      { long SockId, Level, OptName, OptValLen, SockErr;
        unsigned char Result, OptVal[MAXSTRINGLEN];
    
        dbgmsg( "SP.SETSOCKOPT" );
         SockId = get_32(); Level  = get_32();
         OptName = get_32(); OptValLen=get_slice( OptVal);
        dbgmsg( "Socket id %d, Level %d", SockId, Level );
        /*{{{  do setsockopt, sort out error returns*/
        if (setsockopt(SockId,Level,OptName,OptVal,OptValLen)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  getpwdnam()*/
    case SP_GETPWNAM :
      { long SockErr=0, NameLen;
        unsigned char Result, Space[MAXSTRINGLEN], *Name=NULL;
        struct passwd *Pwd;
    
        dbgmsg( "SP.GETPWDNAM" );
        NameLen = get_slice( Space );
        if (NameLen>0) {
          Name=Space;
          dbgmsg( "Name %s",Name );
          }
        if ((Pwd=getpwnam(Name))==NULL) Result=ER_ERROR;
        else Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        /*{{{  transfer data pointed at by Pwd into buffer*/
        if (Result==ER_SUCCESS)
        { int Len, i;
          char *p;
          p=Pwd->pw_name; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_passwd; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          i=Pwd->pw_uid; put_32( i );
          i=Pwd->pw_gid; put_32( i );
          p=Pwd->pw_age; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_comment; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_gecos; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_dir; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_shell; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
        }
        /*}}}*/
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  getpwduid()*/
    case SP_GETPWUID :
      { long MaxPwdLen, SockErr=0, Uid;
        unsigned char Result;
        struct passwd *Pwd;
    
        dbgmsg( "SP.GETPWDUID" );
         Uid  = get_32();
         MaxPwdLen  = get_32();
        dbgmsg( "Uid %d, MaxLen %d",Uid,MaxPwdLen );
        if ((Pwd=getpwuid(Uid))==NULL) Result=ER_ERROR;
        else Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        /*{{{  transfer data pointed at by Pwd into buffer*/
        if (Result==ER_SUCCESS)
        { int Len, i;
          char *p;
	char *TOutBuf=&Tbuf[2];
          p=Pwd->pw_name; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_passwd; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          i=Pwd->pw_uid; put_32( i );
          i=Pwd->pw_gid; put_32( i );
          p=Pwd->pw_age; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_comment; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_gecos; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_dir; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
          p=Pwd->pw_shell; Len=(p!=NULL) ? strlen(p)+1 : 0; put_slice( Len, p );
        }
        /*}}}*/
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  ioctl()*/
    case SP_IOCTL :
      { long SockId, Request, ArgLen, SockErr, Arg, *ArgPtr=NULL, rc;
        unsigned char Result, ArgExist;
        dbgmsg( "SP.IOCTL" );
         SockId = get_32(); Request  = get_32();
         ArgExist =get_8();
        if ((int)ArgExist) {
             Arg  = get_32(); ArgPtr=&Arg;
            }
        dbgmsg( "Socket id %d, Request %8.8x, Arg %8.8x", SockId, Request, Arg );
        /*{{{  do ioctl, sort out error returns*/
        if ((rc=ioctl(SockId,Request,ArgPtr))==-1)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        dbgmsg( "Ioctl returned %d, errno %d, Arg %8.8x", rc, SockErr, Arg );
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if ((int)ArgExist) {
            put_32( Arg );
            }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  listen()*/
    case SP_LISTEN :
      { long SockId, Backlog, SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.LISTEN" );
         SockId = get_32(); Backlog  = get_32();
        dbgmsg( "Socket id %d, Queue length %d", SockId, Backlog );
        /*{{{  do listen, sort out error returns*/
        if (listen(SockId,Backlog)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  read()*/
    case SP_SOCKREAD :
      { long SockId, Length, SockErr, BufOff=0;
        Segment *Seg;
        unsigned char Result;
    
        dbgmsg( "SP.SOCKREAD" );
         SockId = get_32(); Length  = get_32();
        dbgmsg( "Socket id %d, Read length %d", SockId, Length );
        /*{{{  allocate buffer and do read, sort out error returns*/
        if (((Seg=(Segment *)malloc(Length+16))==NULL) ||
            ((Length=read(SockId,Seg->Buff,Length))<0))
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          { long SegAddr=(long)Seg;
            put_32( SegAddr );
            put_32( BufOff );
            put_32( Length );
            Seg->Length=Length; Seg->Done=0;
          }
        else free(Seg);
        dbgmsg( "Actually read %d bytes, allocated at %d", Length, (int)Seg );
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  readv()*/
    case SP_SOCKREADV :
      { long SockId, Length, SockErr, BufOff=0;
        Segment *Seg;
        unsigned char Result;
    
        dbgmsg( "SP.SOCKREADV" );
         SockId = get_32(); Length  = get_32();
        dbgmsg( "Socket id %d, Read length %d", SockId, Length );
        /*{{{  allocate buffer and do read, sort out error returns*/
        if (((Seg=(Segment *)malloc(Length+16))==NULL) ||
            ((Length=read(SockId,Seg->Buff,Length))<0))
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          { long SegAddr=(long)Seg;
            put_32( SegAddr );
            put_32( BufOff );
            put_32( Length );
            Seg->Length=Length; Seg->Done=0;
          }
        else free(Seg);
        dbgmsg( "Actually read %d bytes, allocated at %d", Length, (int)Seg );
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  readsegstart()*/
    case SP_READSEGSTART :
    case SP_RECVITSEGSTART :
      { long SockId, TotLength, Wanted, SendLen, SockErr;
        int FreeIt=0;
        Segment *Seg;
        unsigned char Result, *From;
    
        dbgmsg( "SP.READSEGSTART / SP.RECVITSEGSTART" );
         SockId = get_32(); TotLength = get_32(); Wanted  = get_32();
        dbgmsg( "Socket id %d, Read segment length %d", SockId, Wanted );
        SockErr=0; Seg=(Segment *)SockId;
        dbgmsg( "Done %d, from %d bytes", Seg->Done, Seg->Length );
        if (Seg->Done<=Seg->Length)
          {
            Result=ER_SUCCESS;
            if ((Seg->Done+Wanted)<(Seg->Length)) { SendLen=Wanted; }
            else { SendLen=Seg->Length-Seg->Done; FreeIt=1; }
          }
        else Result=ER_ERROR;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        From=&(Seg->Buff[Seg->Done]); put_slice( SendLen, From );
        (Seg->Done)+=SendLen;
        if (FreeIt) free(Seg);
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  readsegmore()*/
    case SP_READSEGMORE :
    case SP_RECVITSEGMORE :
      { long SockId, Wanted, SendLen, SockErr;
        int FreeIt=0;
        Segment *Seg;
        unsigned char Result, *From;
    
        dbgmsg( "SP.READSEGMORE / SP.RECVITSEGMORE" );
         SockId = get_32(); Wanted  = get_32();
        dbgmsg( "Socket id %d, Read segment length %d", SockId, Wanted );
        SockErr=0; Seg=(Segment *)SockId;
        if (Seg->Done<=Seg->Length)
          {
            Result=ER_SUCCESS;
            if ((Seg->Done+Wanted)<(Seg->Length)) { SendLen=Wanted; }
            else { SendLen=Seg->Length-Seg->Done; FreeIt=1; }
          }
        else Result=ER_ERROR;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        From=&(Seg->Buff[Seg->Done]); put_slice( SendLen, From );
        (Seg->Done)+=SendLen;
        if (FreeIt) free(Seg);
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  recvfrom()*/
    case SP_RECVIT :
      { long SockId, Length, Flags, AddrLen, SockErr, BufOff=0;
        Segment *Seg;
        unsigned char Result, AddrSpace[MAXSTRINGLEN];
        struct sockaddr *Addr=(struct sockaddr *)AddrSpace;
        short Family;
    
        dbgmsg( "SP.RECVFROM" );
         SockId = get_32(); Length  = get_32();
         Flags = get_32(); AddrLen  = get_32();
        if (AddrLen==0) Addr=NULL;
        dbgmsg( "Socket id %d, Read length %d", SockId, Length );
        /*{{{  allocate buffer and do recvfrom, sort out error returns*/
        if (((Seg=(Segment *)malloc(Length+8))==NULL) ||
            ((Length=recvfrom(SockId,Seg->Buff,Length,Flags,Addr,&AddrLen))<0))
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          { long SegAddr=(long)Seg;
            put_32( SegAddr );
            put_32( BufOff );
            put_32( Length );
            Seg->Length=Length; Seg->Done=0;
            if (Addr!=NULL) {
              Family=Addr->sa_family;
              put_16( Family );
              put_slice( AddrLen, AddrSpace );
              }
          }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  select()*/
    case SP_SELECT :
      { long Width, ReadLen, WriteLen, ExceptLen, NoSel, SockErr,
                     *RFds,*WFds,*EFds;
        fd_set ReadFds,WriteFds,ExceptFds;
        struct timeval Timeout, *UseTime; unsigned char Result;
    
        dbgmsg( "SP.SELECT" );
         Width  = get_32();
        /*{{{  get read fds*/
         ReadLen  = get_16();
        if (ReadLen>0)
          { int i;
        
            RFds=(long *)&ReadFds;
            for(i=0;i<ReadLen;i++)
                RFds[i] = get_32();
          }
        else RFds=NULL;
        /*}}}*/
        /*{{{  get write fds*/
         WriteLen  = get_16();
        if (WriteLen>0)
          { int i;
        
            WFds=(long *)&WriteFds;
            for(i=0;i<WriteLen;i++)
                WFds[i] = get_32();
          }
        else WFds=NULL;
        /*}}}*/
        /*{{{  get except fds*/
         ExceptLen  = get_16();
        if (ExceptLen>0)
          { int i;
        
            EFds=(long *)&ExceptFds;
            for(i=0;i<ExceptLen;i++)
                EFds[i] = get_32();
          }
        else EFds=NULL;
        /*}}}*/
         Timeout.tv_sec  = get_32();
         Timeout.tv_usec  = get_32();
        /*{{{  degug info*/
        dbgmsg( "Width %d, Time (secs) %d (usecs) %d", Width, Timeout.tv_sec, Timeout.tv_usec );
        if (RFds!=NULL) dbgmsg( "Read: %8.8x",*RFds );
        if (WFds!=NULL) dbgmsg( "Write: %8.8x",*WFds );
        if (EFds!=NULL) dbgmsg( "Except: %8.8x",*EFds );
        /*}}}*/
        /*{{{  do select, sort out error returns*/
        UseTime=&Timeout;
        if (Timeout.tv_sec==-1) UseTime=NULL;
        if ((NoSel=select(Width,RFds,WFds,EFds,UseTime))<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          {
            put_32( NoSel );
            /*{{{  put read fds*/
            put_16( ReadLen );
            if (RFds!=NULL)
              { int i;
            
                RFds=(long *)&ReadFds;
                for(i=0;i<ReadLen;i++)
                    put_32(RFds[i]);
              }
            /*}}}*/
            /*{{{  put write fds*/
            put_16( WriteLen );
            if (WFds!=NULL)
              { int i;
            
                WFds=(long *)&WriteFds;
                for(i=0;i<WriteLen;i++)
                    put_32(WFds[i]);
              }
            /*}}}*/
            /*{{{  put except fds*/
            put_16( ExceptLen );
            if (EFds!=NULL)
              { int i;
            
                EFds=(long *)&EFds;
                for(i=0;i<ExceptLen;i++)
                    put_32(EFds[i]);
              }
            /*}}}*/
            /*{{{  debug*/
            if (RFds!=NULL) dbgmsg( "Read: %8.8x",*RFds );
            if (WFds!=NULL) dbgmsg( "Write: %8.8x",*WFds );
            if (EFds!=NULL) dbgmsg( "Except: %8.8x",*EFds );
            /*}}}*/
          }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  sendto()*/
    case SP_SENDIT :
      { long SockId, SegAddr, SegOff, Length, Flags, AddrLen, SockErr;
        Segment *Seg;
        unsigned char Result, AddrSpace[MAXSTRINGLEN];
        struct sockaddr *Addr=(struct sockaddr *)AddrSpace;
        short Family;
    
        dbgmsg( "SP.SENDTO" );
         SockId  = get_32();
         SegAddr = get_32(); SegOff  = get_32();
         Length = get_32(); Flags  = get_32();
        AddrLen = get_slice( AddrSpace );
        if (AddrLen>0) { Family  = get_16(); Addr->sa_family=Family;}
        else Addr=NULL;
        dbgmsg( "Socket id %d, Write length %d", SockId, Length );
        Seg=(Segment *)SegAddr;
        /*{{{  write the buffer, sort out error returns*/
        if ((Length=sendto(SockId,Seg->Buff,Length,Flags,Addr,AddrLen))<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
            put_32( Length );
        free(Seg);
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  shutdown()*/
    case SP_SHUTDOWN :
      { long SockId, How, SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.SHUTDOWN" );
         SockId  = get_32();
         How  = get_32();
        dbgmsg( "How %d",How );
        SockErr=0;
        if (shutdown(SockId,How)<0) { Result=ER_ERROR; SockErr=errno; }
        Result=ER_SUCCESS;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  socket()*/
    case SP_SOCKET :
      { long SockId, Domain, Type, Protocol, SockErr;
        unsigned char Result;
    
        dbgmsg( "SP.SOCKET" );
         Domain = get_32(); Type = get_32(); Protocol  = get_32();
        dbgmsg( "Domain %d, Type %d, Protocol %d ", Domain, Type, Protocol );
        /*{{{  do socket, sort out error returns*/
        if ((SockId=socket(Domain,Type,Protocol))<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
        {
          put_32( SockId );
          dbgmsg( "Socket id %d", SockId );
        }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  write()*/
    case SP_SOCKWRITE :
      { long SockId, SegAddr, SegOff, Length, SockErr;
        Segment *Seg;
        unsigned char Result;
    
        dbgmsg( "SP.SOCKWRITE" );
         SockId  = get_32();
         SegAddr = get_32(); SegOff = get_32(); Length  = get_32();
        dbgmsg( "Socket id %d, Write length %d", SockId, Length );
        Seg=(Segment *)SegAddr;
        /*{{{  write the buffer, sort out error returns*/
        if ((Length=write(SockId,Seg->Buff,Length))<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
            put_32( Length );
        else dbgmsg( "Error returned on write, errno=%d",SockErr );
        free(Seg);
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  writev()*/
    case SP_SOCKWRITEV :
      { long SockId, SegAddr, SegOff, Length, SockErr, Iovcnt, i;
        Segment *Seg;
        unsigned char Result;
        struct iovec Iov[50];
    
        dbgmsg( "SP.SOCKWRITEV" );
         SockId  = get_32();
         SegAddr = get_32(); SegOff = get_32(); Length  = get_32();
         Iovcnt  = get_32();
        dbgmsg( "Socket id %d, Write length %d, Iovcnt %d", SockId, Length, Iovcnt );
        if (Iovcnt>50) {
		close_server(1,"Writev : iovec len exceeded");
	}
        Seg=(Segment *)SegAddr;
        /*{{{  compose iovec structure*/
        Iov[0].iov_base=(char *)Seg->Buff;
        for(i=0; i<Iovcnt; i++)
        { int x;
           x  = get_32(); Iov[i].iov_len=x;
          Iov[i+1].iov_base=Iov[i].iov_base+x;
        }
        /*}}}*/
        /*{{{  writev the buffer, sort out error returns*/
        if ((Length=writev(SockId,Iov,Iovcnt))<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
            put_32( Length );
        else dbgmsg( "Error returned on writev, errno=%d",SockErr );
        free(Seg);
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  writesegstart()*/
    case SP_WRITESEGSTART :
    case SP_SENDITSEGSTART :
      { long SockId, TotLength, ThisLen, SockErr;
        Segment *Seg;
        unsigned char Result, *From;
    
        dbgmsg( "SP.WRITESEGSTART" );
         SockId = get_32(); TotLength  = get_32();
        if ((Seg=(Segment *)malloc(TotLength+16))!=NULL)
          {
            Seg->Done=0;
            Seg->Length=TotLength;
            From=&Seg->Buff[Seg->Done];
            ThisLen=((Seg->Done+ThisLen)<=Seg->Length) ? ThisLen : Seg->Length-Seg->Done;
            ThisLen = get_slice( From );
            Seg->Done+=ThisLen;
            dbgmsg( "Socket id %d, Wrote segment length %d", SockId, ThisLen );
            Result=ER_SUCCESS;
          }
        else Result=ER_ERROR;
        SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          { long SegAddr=(long)Seg;
            put_32( ThisLen );
            put_32( SegAddr );
          }
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  writesegmore()*/
    case SP_WRITESEGMORE :
    case SP_SENDITSEGMORE :
      { long SockId, ThisLen, SockErr;
        Segment *Seg;
        unsigned char Result, *From;
    
        dbgmsg( "SP.WRITESEGMORE" );
         SockId  = get_32();
        Seg=(Segment *)SockId;
        ThisLen=((Seg->Done+ThisLen)<=Seg->Length) ? ThisLen : Seg->Length-Seg->Done;
        From=&Seg->Buff[Seg->Done];
         ThisLen = get_slice( From );
        Seg->Done+=ThisLen;
        dbgmsg( "Socket id %d, Wrote segment length %d", SockId, ThisLen );
        Result=ER_SUCCESS; SockErr=0;
        put_8( Result ); put_32( SockErr );
        put_32( OutSigio ); OutSigio=0;
        if (Result==ER_SUCCESS)
          put_32( ThisLen );
        put_count(OutCount); return;
      }
    /*}}}*/
    /*{{{  Unrecognised*/
    default :
      {
        dbgmsg( "UNDEFINED SOCKET COMMAND" );
        put_8( ER_UNIMPLEMENTED );
        put_count(OutCount); return;
      }
    /*}}}*/
    /*}}}*/
    }
}

SpOpenDir()
{
    /*{{{  open()*/
       long SockId, Flags, Mode, SockErr, PathLen;
        unsigned char Result, Path[MAXSTRINGLEN+1];
	DIR *d;
    
         InBuf = &Tbuf[3];
         OutBuf = &Tbuf[2];
         OutCount = 0;
         dbgmsg( "SP.OPENDIR" );
         PathLen = get_slice( Path );
        Path[PathLen<MAXSTRINGLEN ? PathLen : MAXSTRINGLEN]=(char)'\0';
        dbgmsg( "Path %s ", Path );
        /*{{{  do open, sort out error returns*/
        if ((d=opendir(Path))== NULL)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
	    SockId=(long)d;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result );
        if (Result==ER_SUCCESS)
        {
          put_32( SockId );
          dbgmsg( "File id %d", SockId );
        }
        put_count(OutCount); return;
    /*}}}*/
}

SpReadDir()
{
    /*{{{  readdir()*/
       long SockId, Length, SockErr;
        unsigned char Result;
	struct dirent *entryPtr;
    
         InBuf = &Tbuf[3];
         OutBuf = &Tbuf[2];
         OutCount = 0;
         dbgmsg( "SP.READDIR" );
         SockId  = get_32(); 
        dbgmsg( "Socket id %d", SockId );
	if (( entryPtr = readdir((DIR *)SockId)) == NULL )
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result );
        if (Result==ER_SUCCESS)
          { 
		long tempint;
		put_32( entryPtr->d_off );
		put_32( entryPtr->d_fileno );
		tempint = entryPtr->d_reclen;
		put_16( tempint );
		tempint = entryPtr->d_namlen;
		put_16( tempint );
		tempint = entryPtr->d_namlen+1;
		put_slice( tempint, entryPtr->d_name);
          }
        put_count(OutCount); return;
    /*}}}*/
}

SpCloseDir()
{
    /*{{{  close()*/
       long SockId, SockErr;
        unsigned char Result;
    
        InBuf = &Tbuf[3];
        OutBuf = &Tbuf[2];
        OutCount = 0;
        dbgmsg( "SP.CLOSEDIR" );
        SockId  = get_32();
        dbgmsg( "Socket id %d", SockId );
        /*{{{  do close, sort out error returns*/
        if (closedir((DIR *)SockId)<0)
          {
            SockErr=errno;
            Result=ER_ERROR;
          }
        else
          {
            SockErr=0;
            Result=ER_SUCCESS;
          }
        /*}}}*/
        put_8( Result ); 
        put_count(OutCount); return;
    /*}}}*/
}


