/* Copyright INMOS Limited 1988,1990 */

/* CMSIDENTIFIER */
static char *CMS_Id = "PRODUCT:ITEM.VARIANT-TYPE;0(DATE)";

#include <stdio.h>
#include <string.h>

#include "server.h"
#include "iserver.h"
#include "pack.h"
#include "misc.h"

#include "xserver.h"

/*
 *
 *      bool_t pserverext
 *
 *  Pserver extensions are enabled if this is set to TRUE.  If enabled
 *  the server version number returned to the transputer program is
 *  100 more than normal so the transputer runtime knows it can use the
 *  extensions.
 *
 */

#if PSERVER_EXTENSION
bool  pserverext = TRUE;
#endif


/*
 *
 *      miscellaneous variables used only in this module
 *
 */
        bool  terminate = FALSE;      /* terminate server when set to TRUE */
        int  termstat = 0;           /* status returned by exit() */


/*
 *
 *      struct ttimes ttimes
 *
 *  PACT runtime reports system, user and real time spent on root transputer,
 *  which can be displayed by pserver if so desired.
 *
 */

#if PSERVER_EXTENSION
struct ttimes   ttimes = { -1, -1, -1};
#endif


/*
 * SpCommand
 */

void            SpCommand()
{
   bool            All;
   int             Size;
   char           *Cl;
   dbgmsg("SP.COMMAND");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
   All = get_8()?TRUE:FALSE;
   dbgmsg("%d", (int) All);

   if (All == TRUE)
      Cl = RealCommandLine;
   else
      Cl = DoctoredCommandLine;
      
   dbgmsg("command line: %s\n", Cl);

   put_8(ER_SUCCESS);
   Size = strlen(Cl);
   put_slice(Size, Cl);
   put_count(OutCount);
}

/*
 * SpCore
 */

void            SpCore()
{
   long            Offset;
   int             Off, Length;

   dbgmsg("SP.CORE");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

   if (AnalyseSwitch == 0) {
      dbgmsg("not peeked");
      put_8(ER_ERROR);
      put_count(OutCount);
   }
   else {
      Offset = get_32();
      Off = (int) Offset;
      dbgmsg("offset %ld", Offset);

      Length = get_16();
      dbgmsg("length %d", Length);

      if ((Off >= CoreSize) || (Off < 0) || (Length < 0)) {
         put_8(ER_ERROR);
         put_count(OutCount);
         return;
      }

      if (Length + Off > CoreSize) {
         Length = CoreSize - Off;
         put_8(ER_SUCCESS);
         put_slice(Length, &CoreDump[Offset]);
         put_count(OutCount);
      }
      else {
         put_8(ER_SUCCESS);
         put_slice(Length, &CoreDump[Offset]);
         put_count(OutCount);
      }
   }
}

/*
 * SpId
 */

void            SpId()
{
   char            Version, Host, OS, Board;

   dbgmsg("SP.ID");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

#if PSERVER_EXTENSION
   Version = VERSION_ID + (pserverext ? 100 : 0);
#else
   Version = VERSION_ID;
#endif
   Host = HOST_ID;
   OS = OS_ID;
   Board = BOARD_ID;
   
   dbgmsg("version=%d, host=%d, os=%d, board=%d", Version, Host, OS, Board);

   put_8(ER_SUCCESS);
   put_8(Version);
   put_8(Host);
   put_8(OS);
   put_8(Board);
   put_count(OutCount);
}




void            SpUnknown()
{
   infomsg("[Encountered unknown primary tag (%d)]\n", Tbuf[2]);
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
   put_8(ER_UNIMPLEMENTED);
   put_count(OutCount);
}


#if PSERVER_EXTENSION

/*
 *
 *      int server_argc ()
 *
 *  Returns number of command line arguments (argc).
 *  This is a PSERVER extension.
 *
 *  Args:   -                       Rets:   INT32   argc
 *                                          INT32   total-length
 *
 */

SpArgc()
{
   int Count;

   dbgmsg("SP.ARGC");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    /* Return argc and total length and success if pserver extensions   */
    /* enabled.                                                         */
    if (pserverext) {
        int i;
        int len=0;
        char *cmd;
        cmd = DoctoredCommandLine;
        len = strlen(BootFileName)+1 + strlen (cmd) + 1;
        i = 0;Count = 1;
        if ( strlen(cmd) > 0 ){
		while ( i < len ){
		    if ( cmd[i] == ' ' ) Count++;
                    i++;
		}
                Count++;
        }
        dbgmsg( "Command %s %s", BootFileName, cmd );
        dbgmsg( "Count %d Len %d", Count, len );
	put_8( ER_SUCCESS );
        put_32( Count );
        put_32( len );
	put_count(OutCount);
        return;
    }
    /* Return unimplemented if pserver extensions are disabled.         */
    else
        {put_8(ER_UNIMPLEMENTED);put_count(OutCount);return;}
}


/*
 *
 *      int server_argvx ()
 *
 *  Return command line argument argv[x].
 *  This is a PSERVER extension.
 *
 *  Args:   INT32   x               Rets:   VEC     argv[x]
 *
 */

SpArgvx()
{
   unsigned long Length;
   int Count, Index = 0, i;
   char *cmd;

   dbgmsg("SP.ARGVX");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    /* Return argv[x] and success if pserver extensions enabled.        */
    if (pserverext) {
        int     argc;
        char    *argv;

        /* Return dbgmsg if argument number out of range.                */
        argc = get_32();
        cmd = DoctoredCommandLine;
        if ((argc < 0 )||(( argc == 1 )&&( strlen(cmd) < 1 )))
            {put_8(ER_ERROR);put_count(OutCount);return;}
        i = 0;Count = 0; Index = 0;
        if ( argc > 1 ){
            Count = 1;
            while ( i < strlen(cmd) ){
                if ( cmd[i] == ' ' ) Count++;
                i++;
                if ( Count == argc ){ Index = i; break; }
            }
            if (argc > Count)
                {put_8(ER_ERROR);put_count(OutCount);return;}
        }

        put_8(ER_SUCCESS);

        /* Copy argv[x] to sbuffer, set size, and return sbuffer.       */
        if ( argc == 0 ){
            put_slice(strlen(BootFileName), BootFileName);
        }
        else {
            Length = 0; i = Index;
            while (( cmd[i] != ' ' )&&(cmd[i] != '\0')){ Length++;i++; }
            put_slice(Length, &(cmd[Index]));
        }

        /* Return success.                                              */
        put_count(OutCount);return;
    }

    /* Return unimplemented if pserver extensions are disabled.         */
    else
        {put_8(ER_UNIMPLEMENTED);put_count(OutCount);return;}
}


/*
 *
 *      int server_abort ()
 *
 *  Abort server.  As exit, but with special message and return status.
 *  This is a PSERVER extension.
 *
 *  Args:   INT32   stat            Rets:   -
 *
 */

SpAbort()
{

   dbgmsg("SP.ABORT");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    /* Display diagnostic and terminate with special status if enabled. */
    if (pserverext) {

        /* Set terminate flag and return status.                        */
        terminate = TRUE;
        termstat  = get_32();

        /* Print abort message, depending on abort status.              */
        switch (termstat) {
        case 167:   dbgmsg ("transputer aborted");
                    break;
        case 168:   dbgmsg ("transputer aborted - out of memory");
                    break;
        default:    dbgmsg ("transputer aborted with status %d\n", termstat);
                    break;
        }

        /* Succeeded.                                                   */
        put_8(ER_SUCCESS);put_count(OutCount);return;
    }

    /* Return unimplemented if pserver extensions are disabled.         */
    else
        {put_8(ER_UNIMPLEMENTED);put_count(OutCount);return;}
}

/*
 *
 *      int server_times ()
 *
 *  Generated by PACT runtime to tell pserver about times spent in 'user'
 *  and 'system' code, as well as 'real' (= total), to be printed if desired.
 *  This is a PSERVER extension.
 *
 *  Args:   INT32   real            Rets:   -
 *          INT32   user
 *           INT32   sys
 *
 */

SpTimes()
{
   dbgmsg("SP.TIMES");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;

    /* Store user and system times for later use if enabled.            */
    if (pserverext) {

        /* Store times.                                                 */
        ttimes.real = get_32();
        ttimes.user = get_32();
        ttimes.sys  = get_32();

        /* Succeeded.                                                   */
        put_8(ER_SUCCESS);put_count(OutCount);return;
    }

    /* Return unimplemented if pserver extensions are disabled.         */
    else 
        {put_8(ER_UNIMPLEMENTED);put_count(OutCount);return;}
}

/*
 *
 *      int debug_init ()
 *
 *  Dummy debug fn.  Always fails.
 *
 *  Args:   INT32   csdesc              Rets:   -
 *
 */

void SpDebugInit()
{
   dbgmsg("SP.DEBUGINIT");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
#if PSERVER_EXTENSION
    if (pserverext){
        put_8( ER_ERROR );
        put_count( OutCount );
    }
    else
#endif
    put_8( ER_UNIMPLEMENTED );
    put_count( OutCount );
}


/*
 *
 *      int debug_bpt ()
 *
 *  Dummy debug fn.  Always fails.
 *
 *  Args:   -                           Rets:   -
 *
 */

void SpDebugBpt()
{
   dbgmsg("SP.DEBUGBPT");
   InBuf = &Tbuf[3];
   OutBuf = &Tbuf[2];
   OutCount = 0;
#if PSERVER_EXTENSION
    if (pserverext) {
        put_8( ER_ERROR );
        put_count( OutCount );
    }
    else
#endif
    put_8( ER_UNIMPLEMENTED );
    put_count( OutCount );
}



#endif  /* PSERVER_EXTENSION */


/*
 * Eof
 */
