/*{{{  Heads and locals */
/***********************************************/
/* Volvox driver                               */
/* Author : N. Floros                          */
/* Concurrent Computation Group                */
/* University of Southampton                   */
/*					       */
/* 1 November 1991                             */
/* ESPRIT PUMA (P 2701)                        */
/*                                             */
/***********************************************/


#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/transputer.h> /* volvox header file */
#include "linkio.h"

#define NULL_LINK -1

static char DefaultDevice[64] = "/dev/vxv0" ;

static int Active_Link = NULL_LINK;
#define select_width 32 /* number of bits to be checke in each bit mask */
/*}}}*/

/*{{{  Open Link */
int VOLVOXOpenLink( DeviceName )
  char *DeviceName ;
{
	if (Active_Link != NULL_LINK )
		return( -1 );

	if ( ( *DeviceName == 0 ) || (DeviceName == NULL ) )
		strcpy( DeviceName , DefaultDevice) ;

	if ( ( Active_Link = open( "/dev/vxv0", O_RDWR ) ) < 0 )
		return( -1 );
	
	return( Active_Link );
}

/*}}}*/
/*{{{  Close link */
int VOLVOXCloseLink( Link_id )
int Link_id;
{
	if (Link_id != Active_Link )
		return ( -1 );

	close(Active_Link ) ;

	return( 1 ) ;
}

/*}}}*/
/*{{{  reset Link */
#define reset_default  0x10 
int VOLVOXResetLink (Link_id )
int Link_id;
{
	int reset = reset_default ;
	int sig;
	if (Link_id != Active_Link )
		return( -1 );

	if (ioctl(Link_id,VXV_RESET,&reset) == -1 )
	{
		perror("unabel to reset device");
		exit(1) ;
	}
#if 0
	sig = SIGURG ;
	/* set error ansynchronous mode */
	if (ioctl(Link_id,VXV_ERRORASYNC,&sig) == -1)
	{
		perror("unable to set error ansynchronus mode");
		exit(1);
	}
	sig = SIGUSR2;
	/* set read ansynchronous mode */
	if (ioctl(Link_id,VXV_READASYNC,&sig) == -1)
	{
		perror("unable to set error ansynchronous mode");
		exit(1);
	}
#endif
	return( 1 );
}
/*}}}*/
/*{{{  Analyse link */
int VOLVOXAnalyseLink( Link_id)
int Link_id ;
{
	int mask = 0x10 ;
	if (Link_id != Active_Link )
		return( -1 );

/*	ioctl(Link_id, VXV_ANALYSE, &mask ) ;*/

	return(1);
}

/*}}}*/
/*{{{  Test error */
VOLVOXTestError(Link_id)
int Link_id ;
{
	int error_status;

	if (Link_id != Active_Link)
		return(-1);

	ioctl(Link_id,VXV_STAT_IT,&error_status);

	return(error_status);
}

/*}}}*/
/*{{{  Test read */
VOLVOXTestRead(Link_id)
int Link_id ;
{
	if (Link_id != Active_Link)
		return(-1);
	return(1);
}

/*}}}*/
/*{{{  Test write */
VOLVOXTestWrite(Link_id)
int Link_id;
{
	if (Link_id != Active_Link )
		return(-1) ;
	return(1);
}

/*}}}*/
/*{{{  Read Link */
int VOLVOXReadLink(Link_id, Buffer, Count, Time_out)
int Link_id;
char *Buffer;
unsigned int Count;
int Time_out;
{
	int read_n, select_ret,i ;
	int read_mask, write_mask, exception_mask;

	if (Link_id !=Active_Link)
		return(-1);

	if (ioctl(Link_id, VXV_W_TOR, &Time_out) == -1)
	{
		perror("Unable to set the read timeout");
		exit(1);
	}

	read_n = read(Link_id, Buffer, Count);

	return(read_n);
		
}

/*}}}*/
/*{{{  Write link */
/* no timeouts */
int VOLVOXWriteLink(Link_id, Buffer, Count, Time_out)
int Link_id;
char *Buffer;
unsigned int Count;
int Time_out;
{
	int write_n ;

	if (Link_id != Active_Link )
		return(-1);

	if (Count < 1)
		return(-2);


	write_n = write(Link_id,Buffer,Count) ;

	return(write_n);

}
/*}}}*/

